# -*- coding: utf-8 -*-
"""alfredyang@pharbers.com.

This is job template for Pharbers Max Job
"""
import numpy as np
import pandas as pd
from phlogs.phlogs import phlogger
from phs3.phs3 import s3

from pyspark.sql import SparkSession
import time
from pyspark.sql.types import *
from pyspark.sql.types import StringType, IntegerType, DoubleType
from pyspark.sql import functions as func


def execute(max_path, max_path_local, project_name, model_month_left, model_month_right, paths_foradding, test_out_path):
    spark = SparkSession.builder \
        .master("yarn") \
        .appName("sparkOutlier") \
        .config("spark.driver.memory", "1g") \
        .config("spark.executor.cores", "1") \
        .config("spark.executor.instance", "2") \
        .config("spark.executor.memory", "2g") \
        .getOrCreate()
    
    
    phlogger.info('job4_panel')
    
    # 输入
    if project_name == "Sanofi" or project_name == "AZ":
        universe_path = max_path + "/AZ_Sanofi/universe_az_sanofi_base"
        market_path = max_path + u"/AZ_Sanofi/az_sanofi清洗_ma"
        Notarrive_unpublished_paths = paths_foradding.replace(" ","").split(",")
    else:
        universe_path = max_path + "/" + project_name + "/universe_base"
        market_path  = max_path + "/" + project_name + "/mkt_mapping"
    
    raw_data_adding_final_path = test_out_path + "/" + project_name + "/raw_data_adding_final"
    new_hospital_path = max_path_local + "/" + project_name + "/2019new_hospital.xlsx"
    
    
    # 输出
    panel_path = test_out_path + "/" + project_name + "/panel_result"
        
    
    # =========== 数据检查 =============
    phlogger.info('数据检查-start')
    
    # 存储文件的缺失列
    misscols_dict = {}

    # universe file
    universe = spark.read.parquet(universe_path)
    colnames_universe = universe.columns
    misscols_dict.setdefault("universe", [])
    if ("Panel_ID" not in colnames_universe) and ("PHA" not in colnames_universe):
        misscols_dict["universe"].append("Panel_ID/PHA")
    if ("Hosp_name" not in colnames_universe) and ("HOSP_NAME" not in colnames_universe):
        misscols_dict["universe"].append("Hosp_name/HOSP_NAME")
    if "City" not in colnames_universe:
        misscols_dict["universe"].append("City")
    if "Province" not in colnames_universe:
        misscols_dict["universe"].append("Province")
        
    # market file    
    market = spark.read.parquet(market_path)
    colnames_market = market.columns
    misscols_dict.setdefault("market", [])
    if ("标准通用名" not in colnames_market) and ("通用名" not in colnames_market):
        misscols_dict["market"].append("标准通用名")
    if ("model" not in colnames_market) and ("mkt" not in colnames_market):
        misscols_dict["market"].append("model")
        
    # raw_data_adding_final file
    raw_data_adding_final = spark.read.parquet(raw_data_adding_final_path)
    colnames_adding = raw_data_adding_final.columns
    misscols_dict.setdefault("adding_data", [])
    colnamelist = ['PHA', 'City', 'ID', 'Molecule', 'Sales', 'Units', 'Province', 'Month', 'Year', 'min2', 'S_Molecule', 'std_route', 'add_flag']
    # 'min1', 'year_month', 'Brand', 'Form', 'Specifications', 'Pack_Number', 'Manufacturer', 'Source', 'Corp', 'Route', 'Path', 'Sheet', 'BI_hospital_code', 'City_Tier_2010', 
    # "标准商品名", 'S_Molecule_for_gr', 
    for each in colnamelist:
        if each not in colnames_adding:
            misscols_dict["adding_data"].append(each)
    
    # 判断输入文件是否有缺失列
    for eachfile in misscols_dict.keys():
        if len(misscols_dict[eachfile]) == 0:
            del misscols_dict[eachfile]
    # 如果有缺失列，则报错，停止运行
    if misscols_dict:
        phlogger.error('miss columns: %s' % (misscols_dict))
        raise ValueError('miss columns: %s' % (misscols_dict))
        
    phlogger.info('数据检查-Pass')
    
    # =========== 数据执行 =============
    
    phlogger.info('数据执行-start')
 
    # read_universe
    universe = spark.read.parquet(universe_path)
    for col in universe.columns:
        if col in ["City_Tier", "CITYGROUP"]:
            universe = universe.withColumnRenamed(col, "City_Tier_2010")
    universe = universe \
        .withColumnRenamed("Panel_ID", "PHA") \
        .withColumnRenamed("Hosp_name", "HOSP_NAME") \
        .withColumn("City_Tier_2010", universe["City_Tier_2010"].cast(StringType())) \
        .select("PHA", "HOSP_NAME", "Province", "City").distinct()
    universe.persist()
    
    # 读取 market
    market = spark.read.parquet(market_path)
    markets = market.withColumnRenamed("标准通用名", "通用名") \
        .withColumnRenamed("model", "mkt") \
        .select("mkt", "通用名").distinct()
    
    # 读取 raw_data_adding_final
    raw_data_adding_final = spark.read.parquet(raw_data_adding_final_path)
    raw_data_adding_final.persist()
    
    # 生成 panel
    panel = raw_data_adding_final \
        .join(markets, raw_data_adding_final["S_Molecule"] == markets["通用名"], how="left") \
        .drop("Province", "City") \
        .join(universe, on="PHA", how="left") \
        .withColumn("Date", raw_data_adding_final.Year * 100 + raw_data_adding_final.Month)
    
    panel = panel \
        .groupBy("ID", "Date", "min2", "mkt", "HOSP_NAME", "PHA", "S_Molecule", "Province", "City", "add_flag", "std_route") \
        .agg(func.sum("Sales").alias("Sales"), func.sum("Units").alias("Units"))
    
    # 对 panel 列名重新命名
    old_names = ["ID", "Date", "min2", "mkt", "HOSP_NAME", "PHA", "S_Molecule", "Province", "City", "add_flag", "std_route"]
    new_names = ["ID", "Date", "Prod_Name", "DOI", "Hosp_name", "HOSP_ID", "Molecule", "Province", "City", "add_flag", "std_route"]
    for index, name in enumerate(old_names):
        panel = panel.withColumnRenamed(name, new_names[index])
    
    panel = panel \
        .withColumn("Prod_CNAME", panel.Prod_Name) \
        .withColumn("Strength", panel.Prod_Name) \
        .withColumn("DOIE", panel.DOI) \
        .withColumn("Date", panel["Date"].cast(DoubleType()))
    
    # 拆分 panel_raw_data， panel_add_data
    panel_raw_data = panel.where(panel.add_flag == 0)
    panel_raw_data.persist()
    panel_add_data = panel.where(panel.add_flag == 1)
    panel_add_data.persist()
    
    original_Date_molecule = panel_raw_data.select("Date", "Molecule").distinct()
    original_Date_ProdName = panel_raw_data.select("Date", "Prod_Name").distinct()
    
    panel_add_data = panel_add_data \
        .join(original_Date_molecule, on=["Date", "Molecule"], how="inner") \
        .join(original_Date_ProdName, on=["Date", "Prod_Name"], how="inner")
    
    # new_hospital = pd.read_excel(new_hospital_path)
    new_hospital = s3.get_excel_from_s3(new_hospital_path)
    new_hospital = new_hospital["PHA"].tolist()
    
    # 生成 panel_filtered
    # 早于model所用时间（历史数据），用new_hospital补数;
    # 处于model所用时间（模型数据），不补数；
    # 晚于model所用时间（月更新数据），用unpublished和not arrived补数
    if project_name == "Sanofi" or project_name == "AZ":
        kct = [u'北京市', u'长春市', u'长沙市', u'常州市', u'成都市', u'重庆市', u'大连市', u'福厦泉市', u'广州市',
               u'贵阳市', u'杭州市', u'哈尔滨市', u'济南市', u'昆明市', u'兰州市', u'南昌市', u'南京市', u'南宁市', u'宁波市',
               u'珠三角市', u'青岛市', u'上海市', u'沈阳市', u'深圳市', u'石家庄市', u'苏州市', u'太原市', u'天津市', u'温州市',
               u'武汉市', u'乌鲁木齐市', u'无锡市', u'西安市', u'徐州市', u'郑州市', u'合肥市', u'呼和浩特市', u'福州市', u'厦门市',
               u'泉州市', u'珠海市', u'东莞市', u'佛山市', u'中山市']
        city_list = [u'北京市', u'上海市', u'天津市', u'重庆市', u'广州市', u'深圳市', u'西安市', u'大连市', u'成都市', u'厦门市', u'沈阳市']
        Province_list = [u'河北省', u'福建省', u'河北', u"福建"]
    
        panel_add_data = panel_add_data \
            .where(~panel_add_data.City.isin(city_list)) \
            .where(~panel_add_data.Province.isin(Province_list)) \
            .where(~(~(panel_add_data.City.isin(kct)) & (panel_add_data.Molecule == u"奥希替尼")))
    
        # 晚于model所用时间（月更新数据），用unpublished和not arrived补数
        for index, eachfile in enumerate(Notarrive_unpublished_paths):
            if index == 0:
                # Notarrive_unpublished = pd.read_excel(eachfile, dtype=str)
                Notarrive_unpublished = s3.get_excel_from_s3(eachfile, dtype=str)
            else:
                # tmp_file = pd.read_excel(eachfile, dtype=str)
                tmp_file = s3.get_excel_from_s3(eachfile, dtype=str)
                Notarrive_unpublished = Notarrive_unpublished.append(tmp_file)
    
        future_range = spark.createDataFrame(Notarrive_unpublished,
                                             schema=StructType([StructField("ID", StringType(), True),
                                                                StructField("Date", StringType(), True)]))
        future_range = future_range.withColumn("Date", future_range["Date"].cast(DoubleType()))
    
        panel_add_data_future = panel_add_data.where(panel_add_data.Date > int(model_month_right)) \
            .join(future_range, on=["Date", "ID"], how="inner") \
            .select(panel_raw_data.columns)
    
        # 早于model所用时间（历史数据），用new_hospital补数;
        panel_add_data_history = panel_add_data.where(panel_add_data.HOSP_ID.isin(new_hospital)) \
            .where(panel_add_data.Date < int(model_month_left)) \
            .select(panel_raw_data.columns)
    
        panel_filtered = (panel_raw_data.union(panel_add_data_history)).union(panel_add_data_future)
    
    else:
        city_list = [u'北京市', u'上海市', u'天津市', u'重庆市', u'广州市', u'深圳市', u'西安市', u'大连市', u'成都市', u'厦门市', u'沈阳市']
        Province_list = [u'河北省', u'福建省', u'河北', u"福建"]
    
        # 去除 city_list和 Province_list
        panel_add_data = panel_add_data \
            .where(~panel_add_data.City.isin(city_list)) \
            .where(~panel_add_data.Province.isin(Province_list))
        panel_add_data_history = panel_add_data \
            .where(panel_add_data.HOSP_ID.isin(new_hospital)) \
            .where(panel_add_data.Date < int(model_month_left)) \
            .select(panel_raw_data.columns)
        panel_filtered = panel_raw_data.union(panel_add_data_history)
    
    # panel_filtered.groupBy("add_flag").agg({"Sales": "sum"}).show()
    # panel_filtered.groupBy("add_flag").agg({"Sales": "sum"}).show()
    
    panel_filtered = panel_filtered.repartition(2)
    panel_filtered.write.format("parquet") \
        .mode("overwrite").save(panel_path)
        
    phlogger.info("输出 panel_filtered 结果：" + str(panel_path))
     
    phlogger.info('数据执行-Finish')
        
    # =========== 数据验证 =============
    # 与原R流程运行的结果比较正确性: Sanofi与Sankyo测试通过
    if True:
        phlogger.info('数据验证-start')
        
        my_out = spark.read.parquet(panel_path)
        
        if project_name == "Sanofi":
            R_out_path = "/user/ywyuan/max/Sanofi/Rout/panel-result"
        elif project_name == "AZ":
            R_out_path = "/user/ywyuan/max/AZ/Rout/panel-result"
        elif project_name == "Sankyo":
            R_out_path = "/user/ywyuan/max/Sankyo/Rout/panel-result"
            
        R_out = spark.read.parquet(R_out_path)
        
        if project_name == "AZ":
            R_out = R_out.where(R_out.Date/100 < 2020)
            
        # 检查内容：列缺失，列的类型，列的值
        for colname, coltype in R_out.dtypes:
            # 列是否缺失
            if colname not in my_out.columns:
                phlogger.warning ("miss columns:", colname)
            else:
                # 数据类型检查
                if my_out.select(colname).dtypes[0][1] != coltype:
                    phlogger.warning("different type columns: " + colname + ", " + my_out.select(colname).dtypes[0][1] + ", " + "right type: " + coltype)
            
                # 数值列的值检查
                if coltype == "double" or coltype == "int":
                    sum_my_out = my_out.groupBy().sum(colname).toPandas().iloc[0, 0]
                    sum_R = R_out.groupBy().sum(colname).toPandas().iloc[0, 0]
                    # phlogger.info(colname, sum_raw_data, sum_R)
                    if (sum_my_out - sum_R) != 0:
                        phlogger.warning("different value(sum) columns: " + colname + ", " + str(sum_my_out) + ", " + "right value: " + str(sum_R))
        
        phlogger.info('数据验证-Finish')
    

    


# -*- coding: utf-8 -*-
"""alfredyang@pharbers.com.

This is job template for Pharbers Max Job
"""
from phjob import execute
import click


@click.command()
@click.option('--max_path')
@click.option('--max_path_local')
@click.option('--project_name')
@click.option('--model_month_left')
@click.option('--model_month_right')
@click.option('--paths_foradding')
@click.option('--test_out_path')

def debug_execute(max_path, max_path_local, project_name, model_month_left, model_month_right, paths_foradding, test_out_path):
	execute(max_path, max_path_local, project_name, model_month_left, model_month_right, paths_foradding, test_out_path)


if __name__ == '__main__':
    debug_execute()




# -*- coding: utf-8 -*-
"""alfredyang@pharbers.com.

This is job template for Pharbers Max Job
"""
from phjob import execute
import click


@click.command()
@click.option('--max_path')
@click.option('--max_path_local')
@click.option('--project_name')
@click.option('--if_base')
@click.option('--time_left')
@click.option('--time_right')
@click.option('--left_models')
@click.option('--time_left_models')
@click.option('--rest_models')
@click.option('--time_rest_models')
@click.option('--all_models')
@click.option('--other_models')
@click.option('--test_out_path')
def debug_execute(max_path, max_path_local, project_name, if_base, time_left, time_right, left_models, time_left_models, rest_models, time_rest_models, all_models, other_models, test_out_path):
	execute(max_path, max_path_local, project_name, if_base, time_left, time_right, left_models, time_left_models, rest_models, time_rest_models, all_models, other_models, test_out_path)


if __name__ == '__main__':
    debug_execute()




# -*- coding: utf-8 -*-
"""alfredyang@pharbers.com.

This is job template for Pharbers Max Job
"""
import numpy as np
import pandas as pd
from phlogs.phlogs import phlogger


from pyspark.sql import SparkSession
import time
from pyspark.sql.types import *
from pyspark.sql.types import StringType, IntegerType, DoubleType
from pyspark.sql import functions as func


def execute(max_path, max_path_local, project_name, if_base, time_left, time_right, left_models, time_left_models, rest_models, time_rest_models, 
all_models, other_models, test_out_path):
    
    spark = SparkSession.builder \
        .master("yarn") \
        .appName("sparkOutlier") \
        .config("spark.driver.memory", "1g") \
        .config("spark.executor.cores", "1") \
        .config("spark.executor.instance", "2") \
        .config("spark.executor.memory", "2g") \
        .getOrCreate()
    
    phlogger.info('job5_max')
    
    # 输入输出
    if if_base == "F":
        if_base = False
    elif if_base == "T":
        if_base = True
    left_models = left_models.replace(" ","").split(",")
    rest_models = rest_models.replace(" ","").split(",")
    time_parameters = [int(time_left), int(time_right), left_models, int(time_left_models), rest_models, int(time_rest_models)]
    all_models = all_models.replace(" ","").split(",")
    other_models = other_models.replace(" ","").split(",")
    
    if project_name == "Sanofi" or project_name == "AZ":
        project_path = max_path + "/AZ_Sanofi/"
    else:
        project_path = max_path + "/" + project_name
        
    project_path_local = max_path_local + "/" + project_name
    test_out_path = test_out_path + '/' + project_name
        
    # 计算max 函数
    def calculate_max(market, if_base=False, if_box=False):
        
        phlogger.info('market:' + market)
        
        # =========== 输入 =============
        # 根据 market 选择 universe 文件：choose_uni
        if market in ['SNY6', 'SNY10', 'SNY12', 'SNY13', 'AZ12', 'AZ18', 'AZ21']:
            universe_path = project_path + '/universe_az_sanofi_onc'
        elif market in ['SNY5', 'SNY9', 'AZ10', 'AZ11', 'AZ15', 'AZ16', 'AZ14', 'AZ26', 'AZ24']:
            universe_path = project_path + '/universe_az_sanofi_mch'
        else:
            if project_name == "Sanofi" or project_name == "AZ":
                universe_path = project_path + '/universe_az_sanofi_base'
            else:
                universe_path = project_path + '/universe_base'
                
        # universe_outlier_path 以及 factor_path 文件选择
        universe_outlier_path = project_path + "/universe/universe_ot_" + market
        if if_base:
            factor_path = project_path + "/factor/factor_base"
        else:
            factor_path = project_path + "/factor/factor_" + market
            
        # panel 文件选择与读取 获得 original_panel
        if project_name == "Sanofi":
            panel_box_path = project_path + "/panel_box-result_Sanofi"
            panel_path = project_path + "/panel-result_Sanofi"
        elif project_name == "AZ":
            panel_box_path = project_path + "/panel_box-result_AZ"
            panel_path = project_path + "/panel-result_AZ"
        else:
            panel_box_path = project_path + "/panel_box-result"
            panel_path = project_path + "/panel-result"
    
        if if_box:
            original_panel_path = panel_box_path
        else:
            original_panel_path = panel_path
        
        
        # =========== 数据检查 =============
        phlogger.info('数据检查-start')
        
        # 存储文件的缺失列
        misscols_dict = {}
    
        # universe file
        universe = spark.read.parquet(universe_path)
        colnames_universe = universe.columns
        misscols_dict.setdefault("universe", [])
        if ("City_Tier" not in colnames_universe) and ("CITYGROUP" not in colnames_universe) and ("City_Tier_2010" not in colnames_universe):
            misscols_dict["universe"].append("City_Tier/CITYGROUP")
        if ("Panel_ID" not in colnames_universe) and ("PHA" not in colnames_universe):
            misscols_dict["universe"].append("Panel_ID/PHA")
        if ("Hosp_name" not in colnames_universe) and ("HOSP_NAME" not in colnames_universe):
            misscols_dict["universe"].append("Hosp_name/HOSP_NAME")
        if "PANEL" not in colnames_universe:
            misscols_dict["universe"].append("PANEL")
        if "BEDSIZE" not in colnames_universe:
            misscols_dict["universe"].append("BEDSIZE")
        if "Seg" not in colnames_universe:
            misscols_dict["universe"].append("Seg")
            
        # universe_outlier file
        universe_outlier = spark.read.parquet(universe_outlier_path)
        colnames_universe_outlier = universe_outlier.columns
        misscols_dict.setdefault("universe_outlier", [])
        if ("City_Tier" not in colnames_universe) and ("CITYGROUP" not in colnames_universe) and ("City_Tier_2010" not in colnames_universe):
            misscols_dict["universe_outlier"].append("City_Tier/CITYGROUP")
        if ("Panel_ID" not in colnames_universe) and ("PHA" not in colnames_universe):
            misscols_dict["universe_outlier"].append("Panel_ID/PHA")
        if ("Hosp_name" not in colnames_universe) and ("HOSP_NAME" not in colnames_universe):
            misscols_dict["universe_outlier"].append("Hosp_name/HOSP_NAME")
        if "PANEL" not in colnames_universe:
            misscols_dict["universe_outlier"].append("PANEL")
        if "BEDSIZE" not in colnames_universe:
            misscols_dict["universe_outlier"].append("BEDSIZE")
        if "Seg" not in colnames_universe:
            misscols_dict["universe_outlier"].append("Seg")
        if "Est_DrugIncome_RMB" not in colnames_universe:
            misscols_dict["universe_outlier"].append("Est_DrugIncome_RMB")
        
        # factor file
        factor = spark.read.parquet(factor_path)
        colnames_factor = factor.columns
        misscols_dict.setdefault("factor", [])
        if ("factor_new" not in colnames_factor) and ("factor" not in colnames_factor):
            misscols_dict["factor"].append("factor")
        if "City" not in colnames_factor:
            misscols_dict["factor"].append("City")
     
        # original_panel file
        original_panel = spark.read.parquet(original_panel_path)
        colnames_original_panel = original_panel.columns
        misscols_dict.setdefault("original_panel", [])
        colnamelist = ["DOI", 'HOSP_ID', 'Province', 'City', 'Date', 'Molecule', 'Prod_Name', 'Sales', 'Units']    
        for each in colnamelist:
            if each not in colnames_original_panel:
                misscols_dict["original_panel"].append(each)

            
        # 判断输入文件是否有缺失列
        for eachfile in misscols_dict.keys():
            if len(misscols_dict[eachfile]) == 0:
                del misscols_dict[eachfile]
        # 如果有缺失列，则报错，停止运行
        if misscols_dict:
            phlogger.error('miss columns: %s' % (misscols_dict))
            raise ValueError('miss columns: %s' % (misscols_dict))
        
        phlogger.info('数据检查-Pass')
        
        # =========== 数据执行 =============
        
        phlogger.info('数据执行-start')
    
        # 选择 market 的时间范围：choose_months
        time_left = time_parameters[0]
        time_right = time_parameters[1]
        left_models = time_parameters[2]
        time_left_models = time_parameters[3]
        rest_models = time_parameters[4]
        time_rest_models = time_parameters[5]
        
        if market in left_models:
            time_left = time_left_models
        if market in rest_models:
            time_right = time_rest_models
        time_range = str(time_left) + '-' + str(time_right)
    
        # universe_outlier 文件读取与处理：read_uni_ot
        # universe_outlier = spark.read.parquet(universe_outlier_path)
        for col in universe_outlier.columns:
            if col in ["City_Tier", "CITYGROUP"]:
                universe_outlier = universe_outlier.withColumnRenamed(col, "City_Tier_2010")
        universe_outlier = universe_outlier.withColumnRenamed("Panel_ID", "PHA") \
            .withColumnRenamed("Hosp_name", "HOSP_NAME")
        universe_outlier = universe_outlier.withColumn("City_Tier_2010", universe_outlier["City_Tier_2010"].cast(StringType()))
        universe_outlier = universe_outlier.select("PHA", "Est_DrugIncome_RMB", "PANEL", "Seg", "BEDSIZE")
    
        # universe 文件读取与处理：read_universe
        # universe = spark.read.parquet(universe_path)
        for col in universe.columns:
            if col in ["City_Tier", "CITYGROUP"]:
                universe = universe.withColumnRenamed(col, "City_Tier_2010")
        universe = universe.withColumnRenamed("Panel_ID", "PHA") \
            .withColumnRenamed("Hosp_name", "HOSP_NAME")
        universe = universe.withColumn("City_Tier_2010", universe["City_Tier_2010"].cast(StringType()))
    
        # panel 文件读取 获得 original_panel
        # original_panel = spark.read.parquet(original_panel_path)
        original_panel = original_panel.where((original_panel.DOI == market) & (original_panel.Date >= time_left) & (original_panel.Date <= time_right))

    
        # 获得 panel, panel_seg：group_panel_by_seg
    
        # panel：整理成max的格式，包含了所有在universe的panel列标记为1的医院，当作所有样本医院的max
        universe_panel_all = universe.where(universe.PANEL == 1).select('PHA', 'BEDSIZE', 'PANEL', 'Seg')
        panel = original_panel \
            .join(universe_panel_all, original_panel.HOSP_ID == universe_panel_all.PHA, how="inner") \
            .groupBy('PHA', 'Province', 'City', 'Date', 'Molecule', 'Prod_Name', 'BEDSIZE', 'PANEL', 'Seg') \
            .agg(func.sum("Sales").alias("Predict_Sales"), func.sum("Units").alias("Predict_Unit"))
    
        # panel_seg：整理成seg层面，包含了所有在universe_ot的panel列标记为1的医院，可以用来得到非样本医院的max
        panel_drugincome = universe_outlier.where(universe_outlier.PANEL == 1) \
            .groupBy("Seg") \
            .agg(func.sum("Est_DrugIncome_RMB").alias("DrugIncome_Panel"))
        original_panel_tmp = original_panel.join(universe_outlier, original_panel.HOSP_ID == universe_outlier.PHA, how='left')
        panel_seg = original_panel_tmp.where(original_panel_tmp.PANEL == 1) \
            .groupBy('Date', 'Prod_Name', 'Seg', 'Molecule') \
            .agg(func.sum("Sales").alias("Sales_Panel"), func.sum("Units").alias("Units_Panel"))
        panel_seg = panel_seg.join(panel_drugincome, on="Seg", how="left")
    
        # 将非样本的segment和factor等信息合并起来：get_uni_with_factor
        # factor = spark.read.parquet(factor_path)
        if "factor" not in factor.columns:
            factor.withColumnRenamed("factor_new", "factor")
        factor = factor.select('City', 'factor')
        universe_factor_panel = universe.join(factor, on="City", how="left")
        universe_factor_panel = universe_factor_panel \
            .withColumn("factor", func.when(func.isnull(universe_factor_panel.factor), func.lit(1)).otherwise(universe_factor_panel.factor)) \
            .where(universe_factor_panel.PANEL == 0) \
            .select('Province', 'City', 'PHA', 'Est_DrugIncome_RMB', 'Seg', 'BEDSIZE', 'PANEL', 'factor')
    
        # 为这些非样本医院匹配上样本金额、产品、年月、所在segment的drugincome之和
        max_result = universe_factor_panel.join(panel_seg, on="Seg", how="left")
    
        # 预测值等于样本金额乘上当前医院drugincome再除以所在segment的drugincome之和
        max_result = max_result.withColumn("Predict_Sales", (max_result.Sales_Panel / max_result.DrugIncome_Panel) * max_result.Est_DrugIncome_RMB) \
            .withColumn("Predict_Unit", (max_result.Units_Panel / max_result.DrugIncome_Panel) * max_result.Est_DrugIncome_RMB)
    
        # 为什么有空，因为部分segment无样本或者样本金额为0：remove_nega
        max_result = max_result.where(~func.isnull(max_result.Predict_Sales))
        max_result = max_result.withColumn("positive", func.when(max_result["Predict_Sales"] > 0, 1).otherwise(0))
        max_result = max_result.withColumn("positive", func.when(max_result["Predict_Unit"] > 0, 1).otherwise(max_result.positive))
        max_result = max_result.where(max_result.positive == 1).drop("positive")
    
        # 乘上factor
        max_result = max_result.withColumn("Predict_Sales", max_result.Predict_Sales * max_result.factor) \
            .withColumn("Predict_Unit", max_result.Predict_Unit * max_result.factor) \
            .select('PHA', 'Province', 'City', 'Date', 'Molecule', 'Prod_Name', 'BEDSIZE', 'PANEL',
                    'Seg', 'Predict_Sales', 'Predict_Unit')
    
        # 合并样本部分
        max_result = max_result.union(panel.select(max_result.columns))
    
        # 输出结果
        if if_base == False:
            max_result = max_result.repartition(2)
            if if_box:
                max_path = test_out_path + "/MAX_result/MAX_result_" + time_range + market + "_hosp_level_box"
                max_result.write.format("parquet") \
                    .mode("overwrite").save(max_path)
            else:
                max_path = test_out_path + "/MAX_result/MAX_result_" + time_range + market + "_hosp_level"
                max_result.write.format("parquet") \
                    .mode("overwrite").save(max_path)
    
        # 输出excel
        max_excel = max_result.where(max_result.BEDSIZE > 99) \
            .groupBy('Province', 'City', 'PANEL', "Prod_Name", 'Date') \
            .agg(func.sum("Predict_Sales").alias("Predict_Sales"), func.sum("Predict_Unit").alias("Predict_Unit"))
        max_excel = max_excel.toPandas()
        # print(np.sort(max_c["Date"].unique(), axis=None))
    
        if if_box:
            # max_c_path = project_path_local + '/MODEL/' + mkt + '_MAX_result_100bed_' + time_range + '_box' + time.strftime("%Y-%m-%d", time.localtime()) + '.xlsx'
            max_excel_path = project_path_local + '/MODEL/' + market + '_MAX_result_100bed_' + time_range + '_box' + '.xlsx'
            max_excel.to_excel(max_excel_path, index=False)
        else:
            # max_c_path = project_path_local + '/MODEL/' + mkt + '_MAX_result_100bed_' + time_range + time.strftime("%Y-%m-%d", time.localtime()) + '.xlsx'
            max_excel_path = project_path_local + '/MODEL/' + market + '_MAX_result_100bed_' + time_range + '.xlsx'
            max_excel.to_excel(max_excel_path, index=False)
            
        phlogger.info('数据执行-Finish')
    
    
    # 执行函数
    for i in all_models:
        calculate_max(i, if_base=if_base, if_box=False)
    
    for i in other_models:
        calculate_max(i, if_base=if_base, if_box=True)
        
    # =========== 数据验证 =============
    # 与原R流程运行的结果比较正确性:
    if True:
        phlogger.info('数据验证-start')
        
        def check_out(my_out_path, R_out_path):
            my_out = spark.read.parquet(my_out_path)
            R_out = spark.read.parquet(R_out_path)
            
            # 检查内容：列缺失，列的类型，列的值
            for colname, coltype in R_out.dtypes:
                # 列是否缺失
                if colname not in my_out.columns:
                    phlogger.warning ("miss columns:", colname)
                else:
                    # 数据类型检查
                    if my_out.select(colname).dtypes[0][1] != coltype:
                        phlogger.warning("different type columns: " + colname + ", " + my_out.select(colname).dtypes[0][1] + ", " + "right type: " + coltype)
                
                    # 数值列的值检查
                    if coltype == "double" or coltype == "int":
                        sum_my_out = my_out.groupBy().sum(colname).toPandas().iloc[0, 0]
                        sum_R = R_out.groupBy().sum(colname).toPandas().iloc[0, 0]
                        # phlogger.info(colname, sum_raw_data, sum_R)
                        if (sum_my_out - sum_R) != 0:
                            phlogger.warning("different value(sum) columns: " + colname + ", " + str(sum_my_out) + ", " + "right value: " + str(sum_R))
                            
        if project_name == "Sanofi":
            my_out_path = "/user/ywyuan/max/Sanofi/MAX_result/MAX_result_202003-202003SNY6_hosp_level"
            R_out_path = "/user/ywyuan/max/Sanofi/Rout/MAX_result/MAX_result_202003-202003SNY6_hosp_level"
            check_out(my_out_path, R_out_path)
        elif project_name == "AZ":
            # all_models
            my_out_path = "/user/ywyuan/max/AZ/MAX_result/MAX_result_201801-202002SNY5_hosp_level"
            R_out_path = "/user/ywyuan/max/AZ/Rout/MAX_result/MAX_result_201801-202002SNY5_hosp_level"
            phlogger.info("if_box=False:" + str(my_out_path))
            check_out(my_out_path, R_out_path)
            # all_models
            my_out_path = "/user/ywyuan/max/AZ/MAX_result/MAX_result_201701-202002SNY6_hosp_level"
            R_out_path = "/user/ywyuan/max/AZ/Rout/MAX_result/MAX_result_201701-202002SNY6_hosp_level"
            phlogger.info("if_box=False:" + str(my_out_path))
            check_out(my_out_path, R_out_path)
            # other_models
            my_out_path = "/user/ywyuan/max/AZ/MAX_result/MAX_result_201701-202002AZ21_hosp_level_box"
            R_out_path = "/user/ywyuan/max/AZ/Rout/MAX_result/MAX_result_201701-202002AZ21_hosp_level_box"
            phlogger.info("if_box=True:" + str(my_out_path))
            check_out(my_out_path, R_out_path)
        elif project_name == "Sankyo":
            R_out_path = ""
            
        phlogger.info('数据验证-Finish')


from phjob import execute
import click


@click.command()
@click.option('--max_path')
@click.option('--max_path_local')
@click.option('--project_name')
@click.option('--model_month_right')
@click.option('--max_month')
@click.option('--year_missing')
@click.option('--test_out_path')

def debug_execute(max_path, max_path_local, project_name, model_month_right, max_month, year_missing, test_out_path):
    execute(max_path, max_path_local, project_name, model_month_right, max_month, year_missing, test_out_path)


if __name__ == '__main__':
    debug_execute()



# -*- coding: utf-8 -*-
"""alfredyang@pharbers.com.

This is job template for Pharbers Max Job
"""
import numpy as np
import pandas as pd
from phlogs.phlogs import phlogger
from phs3.phs3 import s3

from pyspark.sql import SparkSession
import time
from pyspark.sql.types import *
from pyspark.sql.types import StringType, IntegerType, DoubleType
from pyspark.sql import functions as func


def execute(max_path, max_path_local, project_name, model_month_right, max_month, year_missing, test_out_path):
    spark = SparkSession.builder \
        .master("yarn") \
        .appName("sparkOutlier") \
        .config("spark.driver.memory", "1g") \
        .config("spark.executor.cores", "1") \
        .config("spark.executor.instance", "2") \
        .config("spark.executor.memory", "2g") \
        .getOrCreate()
        

    phlogger.info('job3_data_adding')
    
    # 输入
    product_mapping_out_path = test_out_path + "/" + project_name + "/product_mapping_out"
    products_of_interest_path = max_path_local + "/" + project_name + "/poi.xlsx"
    # model_month_right = 201912
    # project_name = "Sankyo"
    # max_month = 12
    # year_missing = []
    if year_missing:
        year_missing = year_missing.replace(" ","").split(",")
    else:
        year_missing = []
    year_missing = [int(i) for i in year_missing]
    model_month_right = int(model_month_right)
    max_month = int(max_month)

    # 输出
    price_path = test_out_path + "/" + project_name + "/price"
    growth_rate_path = test_out_path + "/" + project_name + "/growth_rate"
    adding_data_path =  test_out_path + "/" + project_name + "/adding_data"
    raw_data_adding_path =  test_out_path + "/" + project_name + "/raw_data_adding"
    new_hospital_path = max_path_local + "/" + project_name + "/2019new_hospital.xlsx"
    raw_data_adding_final_path =  test_out_path + "/" + project_name + "/raw_data_adding_final"

        
    # =========== 数据检查 =============
        
    phlogger.info('数据检查-start')

    # 存储文件的缺失列
    misscols_dict = {}
    
    # product_mapping_out file
    raw_data = spark.read.parquet(product_mapping_out_path)
    colnames_raw_data = raw_data.columns
    misscols_dict.setdefault("product_mapping_out", [])
    
    colnamelist = ['min1', 'PHA', 'City', 'year_month', 'ID',  'Brand', 'Form', 
    'Specifications', 'Pack_Number', 'Manufacturer', 'Molecule', 'Source', 'Corp', 'Route', 
    'Sales', 'Units', 'Path', 'Sheet', 'BI_hospital_code', 'Province', 
    'Month', 'Year', 'City_Tier_2010', 'min2', 'S_Molecule', 'std_route', "标准商品名"]
    #'Raw_Hosp_Name','ORG_Measure','Units_Box'
    
    for each in colnamelist:
        if each not in colnames_raw_data:
            misscols_dict["product_mapping_out"].append(each)
            
    # 判断输入文件是否有缺失列
    for eachfile in misscols_dict.keys():
        if len(misscols_dict[eachfile]) == 0:
            del misscols_dict[eachfile]
    # 如果有缺失列，则报错，停止运行
    if misscols_dict:
        phlogger.error('miss columns: %s' % (misscols_dict))
        raise ValueError('miss columns: %s' % (misscols_dict))        
            
    phlogger.info('数据检查-Pass')        

    # =========== 数据执行 =============
    phlogger.info('数据执行-start')
    
    # 数据读取
    raw_data = spark.read.parquet(product_mapping_out_path)
    raw_data.persist()
    # products_of_interest = pd.read_excel(products_of_interest_path)
    products_of_interest = s3.get_excel_from_s3(products_of_interest_path)
    products_of_interest = products_of_interest["poi"].values.tolist()
    
    # raw_data 处理
    raw_data = raw_data.withColumn("S_Molecule_for_gr",
                                   func.when(raw_data["标准商品名"].isin(products_of_interest), raw_data["标准商品名"]).
                                   otherwise(raw_data.S_Molecule))
                                   
    phlogger.info('1 价格计算')
    
    # 1 价格计算：cal_price 补数部分的数量需要用价格得出
    price = raw_data.groupBy("min2", "year_month", "City_Tier_2010") \
        .agg((func.sum("Sales") / func.sum("Units")).alias("Price"))
    price2 = raw_data.groupBy("min2", "year_month") \
        .agg((func.sum("Sales") / func.sum("Units")).alias("Price2"))
    price = price.join(price2, on=["min2", "year_month"], how="left")
    price = price.withColumn("Price", func.when(func.isnull(price.Price), price.Price2).
                             otherwise(price.Price))
    price = price.withColumn("Price", func.when(func.isnull(price.Price), func.lit(0)).
                             otherwise(price.Price)) \
        .drop("Price2")
    
    # 输出price
    price = price.repartition(2)
    price.write.format("parquet") \
        .mode("overwrite").save(price_path)
    
    phlogger.info("输出 price：" + str(price_path))
    
    # raw_data 处理
    raw_data = raw_data.where(raw_data.Year < ((model_month_right // 100) + 1))
    if project_name == "Sanofi" or project_name == "AZ":
        raw_data = raw_data.where((raw_data.Year > 2016) & (raw_data.Year < 2020))
    
    phlogger.info('2 连续性计算')
    
    
    # 2 计算样本医院连续性: cal_continuity
    # 每个医院每年的月份数
    continuity = raw_data.select("Year", "Month", "PHA").distinct() \
        .groupBy("PHA", "Year").count()
    # 每个医院最大月份数，最小月份数
    continuity_whole_year = continuity.groupBy("PHA") \
        .agg(func.max("count").alias("MAX"), func.min("count").alias("MIN"))
    continuity = continuity.repartition(2, "PHA")
    
    years = continuity.select("Year").distinct().toPandas()["Year"].sort_values().values.tolist()
    # 数据长变宽
    continuity = continuity.groupBy("PHA").pivot("Year").agg(func.sum('count')).fillna(0)
    # 列名修改
    for eachyear in years:
        eachyear = str(eachyear)
        continuity = continuity.withColumn(eachyear, continuity[eachyear].cast(DoubleType())) \
            .withColumnRenamed(eachyear, "Year_" + eachyear)
    # year列求和
    # month_sum = con.Year_2018 + con.Year_2019
    month_sum = ""
    for i in continuity.columns[1:]:
        month_sum += ("continuity." + i + "+")
    month_sum = month_sum.strip('+')
    continuity = continuity.withColumn("total", eval(month_sum))
    # ['PHA', 'Year_2018', 'Year_2019', 'total', 'MAX', 'MIN']
    continuity = continuity.join(continuity_whole_year, on="PHA", how="left")
    
    
    phlogger.info('3 增长率计算')
    
    # 3 计算样本分子增长率: cal_growth
    def calculate_growth(raw_data, max_month=12):
        # TODO: 完整年用完整年增长，不完整年用不完整年增长
        if max_month < 12:
            raw_data = raw_data.where(raw_data.Month <= max_month)
            
        # raw_data 处理
        growth_raw_data = raw_data.na.fill({"City_Tier_2010": 5.0})
        growth_raw_data = growth_raw_data.withColumn("CITYGROUP", growth_raw_data.City_Tier_2010)
    
        # 增长率计算过程
        growth_calculating = growth_raw_data.groupBy("S_Molecule_for_gr", "CITYGROUP", "Year") \
            .agg(func.sum(growth_raw_data.Sales).alias("value"))
    
        years = growth_calculating.select("Year").distinct().toPandas()["Year"].sort_values().values.tolist()
        years = [str(i) for i in years]
        years_name = ["Year_" + i for i in years]
        # 数据长变宽
        growth_calculating = growth_calculating.groupBy("S_Molecule_for_gr", "CITYGROUP").pivot("Year").agg(func.sum('value')).fillna(0)
        growth_calculating = growth_calculating.select(["S_Molecule_for_gr", "CITYGROUP"] + years)
        # 对year列名修改
        for i in range(0, len(years)):
            growth_calculating = growth_calculating.withColumnRenamed(years[i], years_name[i])
    
        # 计算得到年增长： add_gr_cols
        for i in range(0, len(years) - 1):
            growth_rate = growth_calculating.withColumn("GR" + years[i][2:4] + years[i + 1][2:4],
                                                        growth_calculating[years_name[i + 1]] / growth_calculating[years_name[i]])
        # 增长率的调整：modify_gr
        for y in [name for name in growth_rate.columns if name.startswith("GR")]:
            growth_rate = growth_rate.withColumn(y, func.when(func.isnull(growth_rate[y]) | (growth_rate[y] > 10) | (growth_rate[y] < 0.1), 1).
                                                 otherwise(growth_rate[y]))
        return growth_rate
    
    
    # AZ-Sanofi 要特殊处理
    if project_name != "Sanofi" and project_name != "AZ":
        growth_rate = calculate_growth(raw_data)
    else:
        year_missing_df = pd.DataFrame(year_missing, columns=["Year"])
        year_missing_df = spark.createDataFrame(year_missing_df)
        year_missing_df = year_missing_df.withColumn("Year", year_missing_df["Year"].cast(IntegerType()))
        # 完整年
        growth_rate_p1 = calculate_growth(raw_data.join(year_missing_df, on=["Year"], how="left_anti"))
        # 不完整年
        growth_rate_p2 = calculate_growth(raw_data.where(raw_data.Year.isin(year_missing + [y - 1 for y in year_missing] + [y + 1 for y in year_missing])), max_month)
    
        growth_rate = growth_rate_p1.select("S_Molecule_for_gr", "CITYGROUP") \
            .union(growth_rate_p2.select("S_Molecule_for_gr", "CITYGROUP")) \
            .distinct()
        growth_rate = growth_rate.join(
            growth_rate_p1.select(["S_Molecule_for_gr", "CITYGROUP"] + [name for name in growth_rate_p1.columns if name.startswith("GR")]),
            on=["S_Molecule_for_gr", "CITYGROUP"],
            how="left")
        growth_rate = growth_rate.join(
            growth_rate_p2.select(["S_Molecule_for_gr", "CITYGROUP"] + [name for name in growth_rate_p2.columns if name.startswith("GR")]),
            on=["S_Molecule_for_gr", "CITYGROUP"],
            how="left")
    
    # 输出growth_rate结果
    growth_rate = growth_rate.repartition(2)
    growth_rate.write.format("parquet") \
        .mode("overwrite").save(growth_rate_path)
        
    phlogger.info("输出 growth_rate：" + str(growth_rate_path))
    
    phlogger.info('4 补数')
     
    # 4 补数    
    # 4.1 原始数据格式整理， 用于补数: trans_raw_data_for_adding
    growth_rate = growth_rate.select(["CITYGROUP", "S_Molecule_for_gr"] + [name for name in growth_rate.columns if name.startswith("GR1")]) \
        .distinct()
    raw_data_for_add = raw_data.where(raw_data.PHA.isNotNull()) \
        .orderBy(raw_data.Year.desc()) \
        .withColumnRenamed("City_Tier_2010", "CITYGROUP") \
        .join(growth_rate, on=["S_Molecule_for_gr", "CITYGROUP"], how="left")
    raw_data_for_add.persist()
    
    # 4.2 补充各个医院缺失的月份数据:
    
    # add_data
    # 原始数据的 PHA-Month-Year
    original_range = raw_data_for_add.select("Year", "Month", "PHA").distinct()
    original_range.persist()
    
    years = original_range.select("Year").distinct() \
        .orderBy(original_range.Year) \
        .toPandas()["Year"].values.tolist()
    phlogger.info(years)
    
    growth_rate_index = [index for index, name in enumerate(raw_data_for_add.columns) if name.startswith("GR")]
    phlogger.info(growth_rate_index)
    
    # 对每年的缺失数据分别进行补数
    empty = 0
    for eachyear in years:
        # cal_time_range
        # 当前年：月份-PHA
        current_range_pha_month = original_range.where(original_range.Year == eachyear) \
            .select("Month", "PHA").distinct()
        # 当前年：月份
        current_range_month = current_range_pha_month.select("Month").distinct()
        # 其他年：月份-当前年有的月份，PHA-当前年没有的医院
        other_years_range = original_range.where(original_range.Year != eachyear) \
            .join(current_range_month, on="Month", how="inner") \
            .join(current_range_pha_month, on=["Month", "PHA"], how="left_anti")
        # 其他年：与当前年的年份差值，比重计算
        other_years_range = other_years_range \
            .withColumn("time_diff", (other_years_range.Year - eachyear)) \
            .withColumn("weight", func.when((other_years_range.Year > eachyear), (other_years_range.Year - eachyear - 0.5)).
                        otherwise(other_years_range.Year * (-1) + eachyear))
        # 选择比重最小的年份：用于补数的 PHA-Month-Year
        current_range_for_add = other_years_range.repartition(1).orderBy(other_years_range.weight.asc())
        current_range_for_add = current_range_for_add.groupBy("PHA", "Month") \
            .agg(func.first(current_range_for_add.Year).alias("Year"))

    
        # get_seed_data
        # 从 rawdata 根据 current_range_for_add 获取用于补数的数据
        current_raw_data_for_add = raw_data_for_add.where(raw_data_for_add.Year != eachyear) \
            .join(current_range_for_add, on=["Month", "PHA", "Year"], how="inner")
        current_raw_data_for_add = current_raw_data_for_add \
            .withColumn("time_diff", (current_raw_data_for_add.Year - eachyear)) \
            .withColumn("weight", func.when((current_raw_data_for_add.Year > eachyear), (current_raw_data_for_add.Year - eachyear - 0.5)).
                        otherwise(current_raw_data_for_add.Year * (-1) + eachyear))
    
        # cal_seed_with_gr
        # 当前年与(当前年+1)的增长率所在列的index
        base_index = eachyear - min(years) + min(growth_rate_index)
        current_raw_data_for_add = current_raw_data_for_add.withColumn("Sales_bk", current_raw_data_for_add.Sales)
    
        # 为补数计算增长率
        current_raw_data_for_add = current_raw_data_for_add \
            .withColumn("min_index", func.when((current_raw_data_for_add.Year < eachyear), (current_raw_data_for_add.time_diff + base_index)).
                        otherwise(base_index)) \
            .withColumn("max_index", func.when((current_raw_data_for_add.Year < eachyear), (base_index - 1)).
                        otherwise(current_raw_data_for_add.time_diff + base_index - 1)) \
            .withColumn("total_gr", func.lit(1))
    
        for i in growth_rate_index:
            col_name = current_raw_data_for_add.columns[i]
            current_raw_data_for_add = current_raw_data_for_add.withColumn(col_name, func.when((current_raw_data_for_add.min_index > i) | (current_raw_data_for_add.max_index < i), 1).
                                                         otherwise(current_raw_data_for_add[col_name]))
            current_raw_data_for_add = current_raw_data_for_add.withColumn(col_name, func.when(current_raw_data_for_add.Year > eachyear, current_raw_data_for_add[col_name] ** (-1)).
                                                         otherwise(current_raw_data_for_add[col_name]))
            current_raw_data_for_add = current_raw_data_for_add.withColumn("total_gr", current_raw_data_for_add.total_gr * current_raw_data_for_add[col_name])
    
        current_raw_data_for_add = current_raw_data_for_add.withColumn("final_gr", func.when(current_raw_data_for_add.total_gr < 2, current_raw_data_for_add.total_gr).
                                                     otherwise(2))
    
        # 为当前年的缺失数据补数：根据增长率计算 Sales，匹配 price，计算 Units=Sales/price
        current_adding_data = current_raw_data_for_add \
            .withColumn("Sales", current_raw_data_for_add.Sales * current_raw_data_for_add.final_gr) \
            .withColumn("Year", func.lit(eachyear))
        current_adding_data = current_adding_data.withColumn("year_month", current_adding_data.Year * 100 + current_adding_data.Month)
        current_adding_data = current_adding_data.withColumn("year_month", current_adding_data["year_month"].cast(DoubleType()))
    
        current_adding_data = current_adding_data.withColumnRenamed("CITYGROUP", "City_Tier_2010") \
            .join(price, on=["min2", "year_month", "City_Tier_2010"], how="inner")
        current_adding_data = current_adding_data.withColumn("Units", func.when(current_adding_data.Sales == 0, 0).
                                                     otherwise(current_adding_data.Sales / current_adding_data.Price)) \
            .na.fill({'Units': 0})
    
        if empty == 0:
            adding_data = current_adding_data
        else:
            adding_data = adding_data.union(current_adding_data)
        empty = empty + 1
    
    # 测试：输出adding_data
    adding_data = adding_data.repartition(2)
    adding_data.write.format("parquet") \
        .mode("overwrite").save(adding_data_path)
        
    phlogger.info("输出 adding_data：" + str(adding_data_path))
    
    # 1.8 合并补数部分和原始部分:
    # combind_data
    raw_data_adding = (raw_data.withColumn("add_flag", func.lit(0))) \
        .union(adding_data.withColumn("add_flag", func.lit(1)).select(raw_data.columns + ["add_flag"]))
    raw_data_adding.persist()
    
    # 输出
    raw_data_adding = raw_data_adding.repartition(2)
    raw_data_adding.write.format("parquet") \
        .mode("overwrite").save(raw_data_adding_path)
        
    phlogger.info("输出 raw_data_adding：" + str(raw_data_adding_path))
    
    raw_data_for_add.unpersist()
    raw_data.unpersist()
    
    # 1.9 进一步为最后一年独有的医院补最后一年的缺失月（可能也要考虑第一年）:
    
    # 只在最新一年出现的医院
    new_hospital = (original_range.where(original_range.Year == max(years)).select("PHA").distinct()) \
        .subtract(original_range.where(original_range.Year != max(years)).select("PHA").distinct()) \
        .toPandas()
    phlogger.info("以下是最新一年出现的医院:" + str(new_hospital["PHA"].tolist()))
    # 输出
    new_hospital.to_excel(new_hospital_path)
    
    # 最新一年没有的月份
    missing_months = (original_range.where(original_range.Year != max(years)).select("Month").distinct()) \
        .subtract(original_range.where(original_range.Year == max(years)).select("Month").distinct())
    
    # 如果最新一年有缺失月份，需要处理
    if missing_months.count() == 0:
        phlogger.info("missing_months=0")
        raw_data_adding_final = raw_data_adding
    else:
        number_of_existing_months = 12 - missing_months.count()
        # 用于groupBy的列名：raw_data_adding列名去除list中的列名
        group_columns = set(raw_data_adding.columns) \
            .difference(set(['Month', 'Sales', 'Units', '季度', "sales_value__rmb_", "total_units", "counting_units", "year_month"]))
        # 补数重新计算
        adding_data_new = raw_data_adding \
            .where(raw_data_adding.add_flag == 1) \
            .where(raw_data_adding.PHA.isin(new_hospital["PHA"].tolist())) \
            .groupBy(list(group_columns)).agg({"Sales": "sum", "Units": "sum"})
        adding_data_new = adding_data_new \
            .withColumn("Sales", adding_data_new["sum(Sales)"] / number_of_existing_months) \
            .withColumn("Units", adding_data_new["sum(Units)"] / number_of_existing_months) \
            .crossJoin(missing_months)
        # 生成最终补数结果
        same_names = list(set(raw_data_adding.columns).intersection(set(adding_data_new.columns)))
        raw_data_adding_final = raw_data_adding.select(same_names) \
            .union(adding_data_new.select(same_names))
    
    # 输出补数结果 raw_data_adding_final
    raw_data_adding_final = raw_data_adding_final.repartition(2)
    raw_data_adding_final.write.format("parquet") \
        .mode("overwrite").save(raw_data_adding_final_path)
        
    phlogger.info("输出 raw_data_adding_final：" + str(raw_data_adding_final_path))
    
    phlogger.info('数据执行-Finish')
    
    # =========== 数据验证 =============
    # 与原R流程运行的结果比较正确性: Sanofi与Sankyo测试通过
    if True:
        phlogger.info('数据验证-start')
        
        my_out = spark.read.parquet(raw_data_adding_final_path)
        
        if project_name == "Sanofi":
            R_out_path = "/common/projects/max/AZ_Sanofi/adding_data_new"
        elif project_name == "AZ":
            R_out_path = "/user/ywyuan/max/AZ/Rout/adding_data_new"
        elif project_name == "Sankyo":
            R_out_path = "/common/projects/max/Sankyo/adding_data_new"
        R_out = spark.read.parquet(R_out_path)
        
        # 检查内容：列缺失，列的类型，列的值
        for colname, coltype in R_out.dtypes:
            # 列是否缺失
            if colname not in my_out.columns:
                phlogger.warning ("miss columns:", colname)
            else:
                # 数据类型检查
                if my_out.select(colname).dtypes[0][1] != coltype:
                    phlogger.warning("different type columns: " + colname + ", " + my_out.select(colname).dtypes[0][1] + ", " + "right type: " + coltype)
            
                # 数值列的值检查
                if coltype == "double" or coltype == "int":
                    sum_my_out = my_out.groupBy().sum(colname).toPandas().iloc[0, 0]
                    sum_R = R_out.groupBy().sum(colname).toPandas().iloc[0, 0]
                    # phlogger.info(colname, sum_raw_data, sum_R)
                    if (sum_my_out - sum_R) != 0:
                        phlogger.warning("different value(sum) columns: " + colname + ", " + str(sum_my_out) + ", " + "right value: " + str(sum_R))
        
        phlogger.info('数据验证-Finish')
        
    # =========== return =============          
    return raw_data_adding_final
